'''
'''
# Anwar O. Nunez-Elizalde (Apr, 2020)
#
# Updates:
#
#
#
import numpy as np

from moten.core import (mk_spatiotemporal_gabor,
                        generate_3dgabor_array,
                        dotspatial_frames,
                        dotdelay_frames,
                        mk_3d_gabor,
                        )

def plot_3dgabor(vhsize=(576, 1024),
                 gabor_params_dict={},
                 background=None,
                 title=None,
                 speed=1.0, time_padding=False):
    '''Show an animation of the 3D Gabor

    Parameters
    ----------

    '''
    import matplotlib.pyplot as plt
    import matplotlib.animation as animation

    # get the 3D gabor as an array
    spatiotemporal_gabor_components = mk_3d_gabor(vhsize,
                                                  **gabor_params_dict)

    gabor_video = mk_spatiotemporal_gabor(*spatiotemporal_gabor_components)
    assert gabor_video.min() >= -1 and gabor_video.max() <= 1

    ## TODO: wrap plot_3dgabor_array
    # generate individual frames
    fps = gabor_params_dict['stimulus_fps']
    nframes = gabor_video.shape[-1]
    fig, ax = plt.subplots()
    images = []
    for frameidx in range(nframes):
        if background is None:
            image_view = gabor_video[..., frameidx]

        else:
            gmask = np.abs(gabor_video[...,frameidx]) > 0.01
            image_view = background[frameidx].copy()
            image_view[gmask] *= gabor_video[...,frameidx][gmask]

        im = ax.imshow(image_view, vmin=-1, vmax=1,
                       cmap='coolwarm' if background is None else 'Greys')
        framenum = ax.text(0,0,'frame #%04i/%04i (fps=%i)'%(frameidx+1, nframes, fps))
        images.append([framenum, im])


    # create animation
    repeat_delay = 1000*(fps - np.mod(nframes, fps))/fps if time_padding else 0.0
    ani = animation.ArtistAnimation(fig, images,
                                    interval=(1000*(1./fps))*speed,
                                    blit=False,
                                    repeat=True,
                                    repeat_delay=repeat_delay)
    fig.suptitle(title)
    return ani


def plot_3dgabor_array(gabor_video,
                       fps=24, aspect_ratio=1.0, title=None,
                       background=False,
                       speed=1.0, time_padding=False):
    '''
    '''
    import matplotlib.pyplot as plt
    import matplotlib.animation as animation

    # generate individual frames
    nframes = gabor_video.shape[-1]
    fig, ax = plt.subplots()
    images = []
    for frameidx in range(nframes):
        if background is None:
            gabor_view = gabor_video[..., frameidx]
        else:
            gabor_view = background[frameidx]*gabor_video[...,frameidx]

        im = ax.imshow(gabor_view, vmin=-1, vmax=1, cmap='coolwarm')
        framenum = ax.text(0,0,'frame #%04i/%04i (fps=%i)'%(frameidx+1, nframes, fps))
        images.append([framenum, im])


    # create animation
    repeat_delay = 1000*(fps - np.mod(nframes, fps))/fps if time_padding else 0.0
    ani = animation.ArtistAnimation(fig, images,
                                    interval=(1000*(1./fps))*speed,
                                    blit=False,
                                    repeat=True,
                                    repeat_delay=repeat_delay)
    fig.suptitle(title)
    return ani



'''
'''
#
# Adapted from MATLAB code written by S. Nishimoto (see Nishimoto, et al., 2011).
# Anwar O. Nunez-Elizalde (Jan, 2016)
#
# Updates:
#  Anwar O. Nunez-Elizalde (Apr, 2020)


##############################
# internal imports
##############################
from moten import (pyramids,
                   utils,
                   core,
                   viz,
                   io,
                   )


# some default pyramids
default_pyramids = pyramids.DefaultPyramids()


def get_default_pyramid(vhsize=(144, 256), fps=24, **kwargs):
    '''Construct a motion energy pyramid

    A motion energy pyramid consists of a set of
    spatio-temporal Gabor filters that tile the screen.
    motion energy features are extracted by convolving the
    spatio-temporal Gabor filters with the stimulus movie.

    Parameters
    ----------
    vhsize : tuple of ints
        Horizontal and vertical size of the stimulus in [pixels]
    fps : int
        Stimulus frame rate

    Returns
    -------
    pyramid : :class:`moten.pyramids.MotionEnergyPyramid`

    Examples
    --------
    >>> import moten
    >>> video_file = 'http://anwarnunez.github.io/downloads/avsnr150s24fps_tiny.mp4'
    >>> luminance_images = moten.io.video2luminance(video_file, nimages=100)
    >>> nimages, vdim, hdim = luminance_images.shape
    >>> pyramid = moten.get_default_pyramid(vhsize=(vdim, hdim), fps=24)
    >>> moten_features = pyramid.project_stimulus(luminance_images)
    '''
    return pyramids.MotionEnergyPyramid(stimulus_vhsize=vhsize,
                                        stimulus_fps=fps,
                                        **kwargs)


__all__ = ['utils', 'core', 'viz', 'io']

if __name__ == '__main__':
    pass



'''Code to convert RGB to other color spaces
'''
#
# Color transformations from
# /auto/k1/shinji/matlab/colorspace/colorspace2.m
# /auto/k1/shinji/matlab/strflab_adds/preprocColorSpace.m
#
# Code originally by Pascal Getreuer (https://getreuer.info/) 2005-2006:
# https://www.mathworks.com/matlabcentral/fileexchange/28790-colorspace-transformations
#
# Only a few transformations were converted to python.
#
# Anwar O. Nunez-Elizalde (2016).
#
# Updates:
#
#
import numpy as np


def rgb2lab(image):
    '''Convert RGB to CIE LAB color space **in-place**

    Parameters
    ----------
    image : 3D numpy float array
        Array must be in [0,1] range. Last
        dimension corresponds to RGB channels.

    Returns
    -------
    LAB : 3D numpy float array
        The CIE LAB representation in the image.
    '''

    WhitePoint = np.asarray([0.95047,1.0,1.08883])
    # Convert image to XYZ
    image = rgb2xyz(image)
    # % Convert XYZ to CIE L*a*b*
    X = image[:,:,0]/WhitePoint[0]
    Y = image[:,:,1]/WhitePoint[1]
    Z = image[:,:,2]/WhitePoint[2]
    fX = _ff(X)
    fY = _ff(Y)
    fZ = _ff(Z)
    image[:,:,0] = 116.0*fY - 16          # L*
    image[:,:,1] = 500.0*(fX - fY)        # a*
    image[:,:,2] = 200.0*(fY - fZ)        # b*
    return image


def rgb2lch(image):
    '''Convert RGB to CIE LCH color space **in-place**

    Parameters
    ----------
    image : 3D numpy float array
        Array must be in [0,1] range. Last
        dimension corresponds to RGB channels.

    Returns
    -------
    LCH : 3D numpy float array
        The CIE LCH representation in the image.
    '''

    image = rgb2lab(image)
    H = np.arctan2(image[:,:,2],image[:,:,1])
    H = H*180/np.pi + 360*(H < 0.0)
    image[:,:,1] = np.sqrt(image[:,:,1]**2 + image[:,:,2]**2)  # C
    image[:,:,2] = H                                           # H
    return image


def rgb2xyz(image):
    '''Convert RGB to CIE XYZ color space **in-place**

    Parameters
    ----------
    image : 3D numpy float array
        Array must be in [0,1] range. Last
        dimension corresponds to RGB channels.

    Returns
    -------
    LAB : 3D numpy float array
        The CIE XYZ representation in the image.
    '''
    # % Undo gamma correction
    R = _invgammacorrection(image[:,:,0])
    G = _invgammacorrection(image[:,:,1])
    B = _invgammacorrection(image[:,:,2])
    # % Convert RGB to XYZ
    xyz_from_rgb = np.array([[0.412453, 0.357580, 0.180423],
                            [0.212671, 0.715160, 0.072169],
                            [0.019334, 0.119193, 0.950227]])
    T = xyz_from_rgb.T.ravel()
    image[:,:,0] = T[0]*R + T[3]*G + T[6]*B # X
    image[:,:,1] = T[1]*R + T[4]*G + T[7]*B # Y
    image[:,:,2] = T[2]*R + T[5]*G + T[8]*B # Z
    return image


def _ff(Y):
    '''Obscure helper function
    '''
    fY = np.real(Y**(1./3.))
    idx = Y < 0.008856
    fY[idx] = Y[idx]*(7.787) + (4./29.)
    return fY


def _invgammacorrection(Rp):
    '''
    '''
    R = np.real(((Rp + 0.055)/1.055)**(2.4))
    idx = Rp < 0.04045
    R[idx] = Rp[idx] / 12.92
    return R


def _gamma_correct(image, gamma=1.0):
    """Do not use unless you know what you're doing
    """
    if gamma != 1.0:
        if not (0 <= image.min() and 1 >= image.max()):
            image /= 255.
        image = image**gamma
    return image


if __name__ == '__main__':
    pass



'''Compute total motion energy from greyscale videos.
'''
import numpy as np

import moten
from moten.utils import (iterator_func,
                         pointwise_square,
                         )

with_tqdm = False


def pixbypix_covariance_from_frames_generator(data_generator,
                                              batch_size=1000,
                                              output_nonlinearity=pointwise_square,
                                              dtype='float32'):
    '''Compute the pixel-by-pixel covariance from a video frame generator in batches.

    Parameters
    ----------
    data_generator : generator object
        Yields a video frame of shape (vdim, hdim)
    batch_size : optional, int
        Number of frames to process simultaneously while
    output_nonlinearity : optiona, function
        A pointwise function applied to the pixels of each frame.

    Examples
    --------
    >>> import moten
    >>> video_file = 'http://anwarnunez.github.io/downloads/avsnr150s24fps_tiny.mp4'
    >>> small_size = (36, 64)  # downsample to (vdim, hdim) 16:9 aspect ratio
    >>> fdiffgen = moten.io.generate_frame_difference_from_greyvideo(video_file, size=small_size, nimages=333)
    >>> nimages, XTX = moten.extras.pixbypix_covariance_from_frames_generator(fdiffgen) # doctest: +SKIP
    '''
    first_frame = data_generator.__next__()
    vdim, hdim = first_frame.shape
    npixels = vdim*hdim

    framediff_buffer = np.zeros((batch_size, npixels), dtype=dtype)
    XTX = np.zeros((npixels, npixels), dtype=np.float64)
    nframes = 0

    if with_tqdm:
        pbar = tqdm(desc='Computing pixel-by-pixel covariance (p=%i)'%npixels,
                    total=0,
                    unit='[frames]')
    else:
        print('Computing pixel-by-pixel covariance (p=%i)...'%npixels, end='')

    RUN = True
    while RUN:
        framediff_buffer *= 0.0             # clear buffer
        try:
            for batch_frame_idx in range(batch_size):
                frame_difference = data_generator.__next__().reshape(1, -1)
                framediff_buffer[batch_frame_idx] = output_nonlinearity(frame_difference)
        except StopIteration:
            RUN = False
        finally:
            if with_tqdm:
                pbar.update(batch_frame_idx + 1)
            else:
                print('.', end='')

            nframes += batch_frame_idx + 1
            XTX += framediff_buffer.T @ framediff_buffer

    if with_tqdm:
        pbar.close()
    else:
        print('Finished #%i frames'%nframes)

    return nframes, XTX


class StimulusTotalMotionEnergy(object):
    '''Compute the principal components of the total motion energy.

    Total motion energy is defined as the squared difference
    between the previous and current frame. The pixel-by-pixel
    covariance of the total energy is computed frame-by-frame. Then,
    the spatial principal components are estimated. In a second pass,
    the temporal principal components are computed by projecting
    the total energy onto the spatial components.


    Parameters
    ----------
    video_file : str
        Full path to the video file. The video must be greyscale.
    size : optional, tuple (vdim, hdim)
        The desired output image size.
        If specified, the image is scaled or shrunk to this size.
        If not specified, the original size is kept.
    nimages : optional, int
        If specified, only `nimages` frames are loaded.
    batch_size : optional, int
        Number of frames to process simultaneously while
        computing the pixel covariances.
    output_nonlinearity : optional, function
        Defaults to point-wise square.


    Notes
    -----
    The time-by-pixel total motion energy matrix is defined as :math:`T`.
    Its singular value decomposition is :math:`U S V^{\intercal} = T`.
    The spatial components are :math:`V` and the temporal components are :math:`U`.

    As implemented in this class,

    * The spatial components computed are as above (:math:`V`).
    * The temporal compoonents are scaled by their singular values (:math:`US`).
    * The eigenvalues are the squared singular values (:math:`S^2`).


    Attributes
    ----------
    decomposition_spatial_pcs  : np.ndarray, (npixels, npcs)
    decomposition_temporal_pcs : list, (ntimepoints, npcs)
    decomposition_eigenvalues  : np.ndarray, (min(npixels, nframes),)

    Examples
    --------
    >>> import moten.extras
    >>> video_file = 'http://anwarnunez.github.io/downloads/avsnr150s24fps_tiny.mp4'
    >>> small_size = (36, 64) # (vdim, hdim)
    >>> totalmoten = moten.extras.StimulusTotalMotionEnergy(video_file, small_size, nimages=300)
    >>> totalmoten.compute_pixel_by_pixel_covariance()
    >>> totalmoten.compute_spatial_pcs(npcs=10)
    >>> totalmoten.compute_temporal_pcs()
    '''
    def __init__(self,
                 video_file,
                 size=None,
                 nimages=np.inf,
                 batch_size=100,
                 output_nonlinearity=pointwise_square,
                 dtype='float32'):
        '''
        '''
        self.size = size
        self.dtype = dtype
        self.nimages = nimages
        self.video_file = video_file
        self.batch_size = batch_size
        self.output_nonlinearity = output_nonlinearity

    def get_frame_difference_generator(self):
        '''Return a video buffer that generates the difference between
        the previous and the current frame.
        '''
        import moten.io
        generator = moten.io.generate_frame_difference_from_greyvideo(
            self.video_file, size=self.size, nimages=self.nimages, dtype=self.dtype)

        return generator

    def compute_pixel_by_pixel_covariance(self,
                                          generator=None,
                                          ):
        '''Compute the pixel-by-pixel covariance of the total energy video.

        Notes
        -----
        Covariance is estimated in batches.

        Parameters
        ----------
        generator : optional
            The video frame difference generator. Defaults to the ``video_file``
            used to instantiate the class.

        Attributes
        ----------
        covariance_pixbypix : np.ndarray, (npixels, npixels)
            The full covarianc matrix
        covariance_nframes : int
            Number of frames used in estimating the covariance matrix.
            Defaults to the total number of frames in the video.
        npixels : int
            Total number of pixels in the video (after downsampling).
        '''
        import moten.utils

        if generator is None:
            # allow the user to provide their own frame difference generator
            generator = self.get_frame_difference_generator()

        nframes, xtx = pixbypix_covariance_from_frames_generator(
            generator, batch_size=self.batch_size,
            dtype=self.dtype, output_nonlinearity=self.output_nonlinearity)
        self.covariance_pixbypix = xtx
        self.covariance_nframes = nframes
        self.npixels = xtx.shape[0]

    def compute_spatial_pcs(self, npcs=None):
        '''Compute the principal components from the pixel-by-pixel total energy covariance matrix.

        Parameters
        ----------
        npcs : optional, int
            Number of principal components to keep

        Attributes
        ----------
        decomposition_spatial_pcs : np.ndarray, (npixels, npcs)
        decomposition_eigenvalues : np.ndarray
        '''
        from scipy import linalg

        if npcs is None:
            npcs = min(self.npixels, self.covariance_nframes) + 1

        # Recall the eigendecomposition
        # Q L QT = XTX
        # U,S,Vt = X
        # Q = V
        L, Q = linalg.eigh(self.covariance_pixbypix)

        # increasing order
        L = L[::-1]             # eigenvals (npcs)
        Q = Q[:, ::-1]          # eigenvecs (npixels, npcs)

        # store: (npixels, npcs)
        self.decomposition_spatial_pcs = np.asarray(Q[:, :npcs], dtype=self.dtype)
        self.decomposition_eigenvalues = np.asarray(L, dtype=self.dtype)

    def compute_temporal_pcs(self, generator=None, skip_first=False):
        '''Extract the temporal principal components of the total motion energy.

        Parameters
        ----------
        generator : optional
            The video frame difference generator. Defaults to the ``video_file``
            used to instantiate the class.
        skip_first : optional, bool
            By default, set the first timepoint of all the PCs is set to zeros because
            the first timepoint corresponds to the difference between the first frame and nothing.
            If `skip_first=True`, then the first frame is removed from the timecourse.

        Attributes
        ----------
        decomposition_temporal_pcs : list, (ntimepoints, npcs)
            The temporal compoonents are scaled by their singular values (:math:`US`).
        '''
        import moten.utils

        if generator is None:
            # allow the user to provide their own frame difference generator
            generator = self.get_frame_difference_generator()

        if skip_first:
            # drop the first frame b/c the difference is with 0's
            # and so projection is with itself
            generator.__next__()

        self.decomposition_temporal_pcs = []
        ## TODO: batch for faster performance
        for frame_diff in iterator_func(generator,
                                             'projecting stimuli',
                                             unit='[frames]'):
            # flatten and square
            frame_diff = self.output_nonlinearity(frame_diff.reshape(1, -1))
            projection = frame_diff @ self.decomposition_spatial_pcs
            self.decomposition_temporal_pcs.append(projection.squeeze())

        if skip_first is False:
            # if not skipped, set the first frame to 0s
            # b/c its difference is not really defined
            self.decomposition_temporal_pcs[0][:] *= 0



'''Help functions.
'''
# Anwar O. Nunez-Elizalde (Jan, 2016)
#
# Updates:
#  Anwar O. Nunez-Elizalde (Apr, 2020)
#
#
import numpy as np

with_tqdm = True
try:
    from tqdm import tqdm
except ImportError:
    with_tqdm = False
    pass


##############################
# helper functions
##############################

def log_compress(x, offset=1e-05):
    '''Computes :math:`log(x + \epsilon)`
    '''
    return np.log(x + offset)


def sqrt_sum_squares(x,y):
    '''Computes :math:`(x^2 + y^2)^{1/2}`
    '''
    return np.sqrt(x**2 + y**2)

def pointwise_square(data):
    '''Computes :math:`x^2`
    '''
    return data**2


class DotDict(dict):
    '''A dictionary with dot syntax.
    '''
    # Copied from pykilosort (written by C. Rossant).
    def __init__(self, *args, **kwargs):
        super(type(self), self).__init__(*args, **kwargs)
        self.__dict__ = self

    def copy(self):
        """
        """
        return DotDict(super(type(self), self).copy())


def iterator_func(*args, **kwargs):
    '''If available, show iteration progress with `tqdm`.
    '''
    try:
        from tqdm import tqdm
        return tqdm(*args, **kwargs)
    except ImportError:
        return args[0]
    raise ValueError('Unknown')



'''
'''
# Anwar O. Nunez-Elizalde (Jan, 2016)
#
# Updates:
#  Anwar O. Nunez-Elizalde (Apr, 2020)
#
#
from PIL import Image
import numpy as np

from moten.utils import iterator_func
from moten.colorspace import rgb2lab

def video_buffer(video_file, nimages=np.inf):
    '''Generator for a video file.

    Yields individual uint8 images from a video file.
    The video is loaded into memory one frame at a time.

    Parameters
    ----------
    video_file : str
        Full path to the video file.
        This can be a video file on disk or from a website.
    nimages : optional, int
        If specified, only `nimages` frames are loaded.

    Yields
    ------
    video_frame : uint8 3D np.ndarray, (vdim, hdim, color)
        Each ``next()`` call yields an uint8 RGB frame from video.

    Example
    -------
    >>> video_file = 'http://anwarnunez.github.io/downloads/avsnr150s24fps_tiny.mp4'
    >>> image_buffer = video_buffer(video_file, nimages=50)
    >>> movie = np.asarray([frame for frame in image_buffer])
    >>> print(movie.shape) # (nimages, vdim, hdim, color)
    (50, 144, 256, 3)
    '''
    import cv2
    cap = cv2.VideoCapture(video_file)
    frameidx = 0
    while True:
        if frameidx >= nimages:
            break
        flag, im = cap.read()
        frameidx += 1
        if flag:
            yield im[...,::-1] # flip to RGB
        else:
            break


def generate_frames_from_greyvideo(video_file, size=None, nimages=np.inf):
    '''Yields one frame from the greyscale video file.

    Notes
    -----
    The video is assumed to be greyscale.

    Parameters
    ----------
    video_file : str
        Full path to the video file.
        This can be a video file on disk or from a website.
    size : optional, tuple (vdim, hdim)
        The desired output image size.
        If specified, the image is scaled or shrunk to this size.
        If not specified, the original size is kept.
    nimages : optional, int
        If specified, only `nimages` frames are loaded.

    Yields
    ------
    greyscale_image : 2D np.ndarray, (vdim, hdim)
        Pixel values are in the 0-1 range.
    '''
    vbuffer = video_buffer(video_file, nimages=nimages)
    grey_video = []
    for imageidx, image_rgb in enumerate(vbuffer):
        if size is not None:
            image_rgb = resize_image(image_rgb, size=size)
        grey_image = image_rgb[...,0]/255.0
        yield grey_image


def generate_frame_difference_from_greyvideo(video_file,
                                             size=None,
                                             nimages=np.inf,
                                             dtype='float32'):
    '''Generates the difference between the current frame and the previous frame.

    Notes
    -----
    The video is assumed to be greyscale.

    Parameters
    ----------
    video_file : str
        Full path to the video file.
        This can be a video file on disk or from a website.
    size : optional, tuple (vdim, hdim)
        The desired output image size.
        If specified, the image is scaled or shrunk to this size.
        If not specified, the original size is kept.
    nimages : optional, int
        If specified, only `nimages` frames are loaded.

    Yields
    ------
    greyscale_image_difference : 2D np.ndarray, (vdim, hdim)
        The difference image: (current_frame - previous_frame).
        Pixel values are in the (-1, 1) range.
    '''
    vbuffer = video_buffer(video_file, nimages=nimages)
    previous_frame = 0
    for frame_index, image_rgb in enumerate(vbuffer):
        if size is not None:
            image_rgb = resize_image(image_rgb, size=size)
        current_image = np.asarray(image_rgb[...,0]/255.0, dtype=dtype)
        frame_difference = current_image - previous_frame
        previous_frame = current_image
        yield frame_difference


def video2luminance(video_file, size=None, nimages=np.inf):
    '''Convert the video frames to luminance images.

    Internally, this function loads one video frame into memory at a time.
    Tt converts the RGB pixel values from one frame to CIE-LAB pixel values.
    It then keeps the luminance channel only. This process is performed
    for all frames requested or until we reach the end of the video file.

    Parameters
    ----------
    video_file : str
        Full path to the video file.
        This can be a video file on disk or from a website.
    size : optional, tuple (vdim, hdim)
        The desired output image size.
        If specified, the image is scaled or shrunk to this size.
        If not specified, the original size is kept.
    nimages : optional, int
        If specified, only `nimages` frames are converted to luminance.

    Returns
    -------
    luminance_images : 3D np.ndarray, (nimages, vdim, hdim)
        Pixel values are in the 0-100 range.
    '''
    vbuffer = video_buffer(video_file, nimages=nimages)
    luminance_video = []
    for imageidx, image_rgb in iterator_func(enumerate(vbuffer),
                                             '%s.video2luminance'%__name__):
        luminance_image = imagearray2luminance(image_rgb, size=size).squeeze()
        luminance_video.append(luminance_image)
    return np.asarray(luminance_video)


def video2grey(video_file, size=None, nimages=np.inf):
    '''Convert the video frames to greyscale images.

    This function computes the mean across RGB color channels.

    Parameters
    ----------
    video_file : str
        Full path to the video file.
        This can be a video file on disk or from a website.
    size : optional, tuple (vdim, hdim)
        The desired output image size.
        If specified, the image is scaled or shrunk to this size.
        If not specified, the original size is kept.
    nimages : optional, int
        If specified, only `nimages` frames are converted to greyscale.

    Returns
    -------
    greyscale_images : 3D np.ndarray, (nimages, vdim, hdim)
        Pixel values are in the 0-1 range.
    '''
    vbuffer = video_buffer(video_file, nimages=nimages)
    grey_video = []
    for imageidx, image_rgb in iterator_func(enumerate(vbuffer),
                                             '%s.video2grey'%__name__):
        if size is not None:
            image_rgb = resize_image(image_rgb, size=size)
        grey_image = image_rgb.mean(-1)/255.0
        grey_video.append(grey_image)
    return np.asarray(grey_video)


def imagearray2luminance(uint8arr, size=None, filter=Image.ANTIALIAS, dtype=np.float64):
    '''Convert an array of uint8 RGB images to a luminance image

    Parameters
    ----------
    uint8arr : 4D np.ndarray, (nimages, vdim, hdim, color)
        The uint8 RGB frames.
    size : optional, tuple (vdim, hdim)
        The desired output image size.
    filter: to be passed to PIL

    Returns
    -------
    luminance_array : 3D np.ndarray, (nimages, vdim, hdim)
        The luminance image representation.
        Pixel values are in the 0-100 range.
    '''
    from scipy import misc
    from moten.colorspace import rgb2lab

    if uint8arr.ndim == 3:
        # handle single image case
        uint8arr = np.asarray([uint8arr])

    nimages, vdim, hdim, cdim = uint8arr.shape
    outshape = (nimages, vdim, hdim) if size is None \
        else (nimages, size[0], size[1])

    luminance = np.zeros(outshape, dtype=dtype)
    for imdx in range(nimages):
        im = uint8arr[imdx]
        if size is not None:
            im = Image.fromarray(im)
            im = resize_image(im, size=size, filter=filter)
        im = rgb2lab(im/255.)[...,0]
        luminance[imdx] = im
    return luminance


def resize_image(im, size=(96,96), filter=Image.ANTIALIAS):
    '''Resize an image and return its array representation.

    Parameters
    ----------
    im : str, np.ndarray(uint8), or PIL.Image object
        The path to the image, an image array, or a loaded PIL.Image.
    size : tuple, (vdim, hdim)
        The desired output image size.

    Returns
    -------
    arr : uint8 np.array, (vdim, hdim, 3)
        The resized image array
    '''
    if isinstance(im, str):
        im = Image.open(im)
    elif isinstance(im, np.ndarray):
        im = Image.fromarray(im)
    im.load()

    # flip to PIL.Image convention
    size = size[::-1]
    try:
        im = im._new(im.im.stretch(size, filter))
    except AttributeError:
        # PIL 4.0.0 The stretch function on the core image object has been removed.
        # This used to be for enlarging the image, but has been aliased to resize recently.
        im = im._new(im.im.resize(size, filter))
    im = np.asarray(im)
    return im


def load_image_luminance(image_files, hdim=None, vdim=None):
    '''Load a set of RGB images and return its luminance representation.

    Parameters
    ----------
    image_files : list-like, (nimages,)
        A list of file names.
        The images should be in RGB uint8 format.
    vdim : int, optional
    hdim : int, optional
        Vertical and horizontal dimensions, respectively.
        If provided the images will be scaled to this size.

    Returns
    -------
    arr : 3D np.array (nimages, vdim, hdim)
        The luminance representation of the images.
    '''
    if (hdim and vdim):
        loader = lambda stim,sz: resize_image(stim,sz)
    else:
        loader = lambda stim,sz: np.asarray(stim)

    stimuli = []

    for fdx, fl in iterator_func(enumerate(image_files),
                                 "load_image_luminance",
                                 total=len(image_files)):
        stimulus = Image.open(fl)
        stimulus = loader(stimulus,(vdim,hdim))
        stimulus = rgb2lab(stimulus/255.)[...,0]
        stimuli.append(stimulus)
    return np.asarray(stimuli)



'''
'''
#
# Adapted from MATLAB code written by S. Nishimoto (see Nishimoto, et al., 2011).
# Anwar O. Nunez-Elizalde (Jan, 2016)
#
# Updates:
#  Anwar O. Nunez-Elizalde (Apr, 2020)
#
import itertools
from PIL import Image

import numpy as np

from moten.utils import (DotDict,
                         iterator_func,
                         log_compress,
                         sqrt_sum_squares,
                         pointwise_square,
                         )


##############################
#
##############################

def raw_project_stimulus(stimulus,
                         filters,
                         vhsize=(),
                         dtype='float32'):
    '''Obtain responses to the stimuli from all filter quadrature-pairs.

    Parameters
    ----------
    stimulus : np.ndarray, (nimages, vdim, hdim) or (nimages, npixels)
        The movie frames.
        If `stimulus` is two-dimensional with shape (nimages, npixels), then
        `vhsize=(vdim,hdim)` is required and `npixels == vdim*hdim`.

    Returns
    -------
    output_sin : np.ndarray, (nimages, nfilters)
    output_cos : np.ndarray, (nimages, nfilters)
    '''
    # parameters
    if stimulus.ndim == 3:
        nimages, vdim, hdim = stimulus.shape
        stimulus = stimulus.reshape(stimulus.shape[0], -1)
        vhsize = (vdim, hdim)

    # checks for 2D stimuli
    assert stimulus.ndim == 2                             # (nimages, pixels)
    assert isinstance(vhsize, tuple) and len(vhsize) == 2 # (hdim, vdim)
    assert np.product(vhsize) == stimulus.shape[1]        # hdim*vdim == pixels

    # Compute responses
    nfilters = len(filters)
    nimages = stimulus.shape[0]
    sin_responses = np.zeros((nimages, nfilters), dtype=dtype)
    cos_responses = np.zeros((nimages, nfilters), dtype=dtype)

    for gaborid, gabor_parameters in iterator_func(enumerate(filters),
                                                   'project_stimulus',
                                                   total=len(filters)):

        sgabor0, sgabor90, tgabor0, tgabor90 = mk_3d_gabor(vhsize, **gabor_parameters)

        channel_sin, channel_cos = dotdelay_frames(sgabor0, sgabor90,
                                                   tgabor0, tgabor90,
                                                   stimulus)

        sin_responses[:, gaborid] = channel_sin
        cos_responses[:, gaborid] = channel_cos

    return sin_responses, cos_responses


def project_stimulus(stimulus,
                     filters,
                     quadrature_combination=sqrt_sum_squares,
                     output_nonlinearity=log_compress,
                     vhsize=(),
                     dtype='float32'):
    '''Compute the motion energy filter responses to the stimuli.

    Parameters
    ----------
    stimulus : np.ndarray, (nimages, vdim, hdim) or (nimages, npixels)
        The movie frames.
        If `stimulus` is two-dimensional with shape (nimages, npixels), then
        `vhsize=(vdim,hdim)` is required and `npixels == vdim*hdim`.

    Returns
    -------
    filter_responses : np.ndarray, (nimages, nfilters)
    '''
    # parameters
    if stimulus.ndim == 3:
        nimages, vdim, hdim = stimulus.shape
        stimulus = stimulus.reshape(stimulus.shape[0], -1)
        vhsize = (vdim, hdim)

    # checks for 2D stimuli
    assert stimulus.ndim == 2                             # (nimages, pixels)
    assert isinstance(vhsize, tuple) and len(vhsize) == 2 # (hdim, vdim)
    assert np.product(vhsize) == stimulus.shape[1]        # hdim*vdim == pixels

    # Compute responses
    nfilters = len(filters)
    nimages = stimulus.shape[0]
    filter_responses = np.zeros((nimages, nfilters), dtype=dtype)
    for gaborid, gabor_parameters in iterator_func(enumerate(filters),
                                                   'project_stimulus',
                                                   total=len(filters)):

        sgabor0, sgabor90, tgabor0, tgabor90 = mk_3d_gabor(vhsize, **gabor_parameters)

        channel_sin, channel_cos = dotdelay_frames(sgabor0, sgabor90,
                                                   tgabor0, tgabor90,
                                                   stimulus)

        channel_response = quadrature_combination(channel_sin, channel_cos)
        channel_response = output_nonlinearity(channel_response)
        filter_responses[:, gaborid] = channel_response
    return filter_responses


##############################
# core functionality
##############################

def mk_3d_gabor(vhsize,
                stimulus_fps,
                aspect_ratio='auto',
                filter_temporal_width='auto',
                centerh=0.5,
                centerv=0.5,
                direction=45.0,
                spatial_freq=16.0,
                spatial_env=0.3,
                temporal_freq=2.0,
                temporal_env=0.3,
                spatial_phase_offset=0.0,
                ):
    '''Make a motion energy filter.

    A motion energy filter is a 3D gabor with
    two spatial and one temporal dimension.
    Each dimension is defined by two sine waves which
    differ in phase by 90 degrees. The sine waves are
    then multiplied by a gaussian.

    Parameters
    ----------
    vhsize : tuple of ints,  (vdim, hdim)
        Size of the stimulus in pixels (vdim, hdim)
        `vdim` : vertical dimension
        `hdim` : horizontal dimension
    stimulus_fps : scalar, [Hz]
        Stimulus playback speed in frames per second.
    centerv : scalar
        Vertical filter position from top of frame (min=0, max=1.0).
    centerh : scalar
        Horizontal filter position from left of frame (min=0, max=aspect_ratio).
    direction : scalar, [degrees]
        Direction of filter motion. Degree position corresponds
        to standard unit-circle coordinates (i.e. 0=right, 180=left).
    spatial_freq : float, [cycles-per-image]
        Spatial frequency of the filter.
    temporal_freq : float, [Hz]
        Temporal frequency of the filter
    filter_temporal_width : int
        Temporal window of the motion energy filter (e.g. 10).
        Defaults to approximately 0.666[secs] (`floor(stimulus_fps*(2/3))`).
    aspect_ratio : optional, 'auto' or float-like,
        Defaults to stimulus aspect ratio: hdim/vdim
        Useful for preserving the spatial gabors circular even
        when images have non-square aspect ratios. For example,
        a 16:9 image would have `aspect_ratio`=16/9.

    spatial_env : float
        Spatial envelope (s.d. of the gaussian)
    temporal_env : float
        Temporal envelope (s.d. of gaussian)
    spatial_phase_offset : float, [degrees
        Phase offset for the spatial sinusoid

    Returns
    -------
    spatial_gabor_sin : 2D np.ndarray, (vdim, hdim)
    spatial_gabor_cos : 2D np.ndarray, (vdim, hdim)
        Spatial gabor quadrature pair. ``spatial_gabor_cos`` has
        a 90 degree phase offset relative to ``spatial_gabor_sin``

    temporal_gabor_sin : 1D np.ndarray, (`filter_temporal_width`,)
    temporal_gabor_cos : 1D np.ndarray, (`filter_temporal_width`,)
        Temporal gabor quadrature pair. ``temporal_gabor_cos`` has
        a 90 degree phase offset relative to ``temporal_gabor_sin``

    Notes
    -----
    Same method as Nishimoto, et al., 2011.
    '''
    vdim, hdim = vhsize
    if aspect_ratio == 'auto':
        aspect_ratio = hdim/float(vdim)

    if filter_temporal_width == 'auto':
        filter_temporal_width = int(stimulus_fps*(2/3.))

    # cast filter width to integer frames
    assert np.allclose(filter_temporal_width, int(filter_temporal_width))
    filter_temporal_width = int(filter_temporal_width)

    dh = np.linspace(0, aspect_ratio, hdim, endpoint=True)
    dv = np.linspace(0, 1, vdim, endpoint=True)
    dt = np.linspace(0, 1, filter_temporal_width, endpoint=False)
    # AN: Actually, `dt` should include endpoint.
    # Currently, the center of the filter width is +(1./fps)/2.
    # However, this would break backwards compatibility.
    # TODO: Allow for `dt_endpoint` as an argument
    # and set default to False.

    ihs, ivs = np.meshgrid(dh,dv)

    fh = -spatial_freq*np.cos(direction/180.*np.pi)*2*np.pi
    fv = spatial_freq*np.sin(direction/180.*np.pi)*2*np.pi
    # normalize temporal frequency to wavelet size
    ft = np.real(temporal_freq*(filter_temporal_width/float(stimulus_fps)))*2*np.pi

    # spatial filters
    spatial_gaussian = np.exp(-((ihs - centerh)**2 + (ivs - centerv)**2)/(2*spatial_env**2))

    spatial_grating_sin = np.sin((ihs - centerh)*fh + (ivs - centerv)*fv + spatial_phase_offset)
    spatial_grating_cos = np.cos((ihs - centerh)*fh + (ivs - centerv)*fv + spatial_phase_offset)

    spatial_gabor_sin = spatial_gaussian * spatial_grating_sin
    spatial_gabor_cos = spatial_gaussian * spatial_grating_cos

    ##############################
    temporal_gaussian = np.exp(-(dt - 0.5)**2/(2*temporal_env**2))
    temporal_grating_sin = np.sin((dt - 0.5)*ft)
    temporal_grating_cos = np.cos((dt - 0.5)*ft)

    temporal_gabor_sin = temporal_gaussian*temporal_grating_sin
    temporal_gabor_cos = temporal_gaussian*temporal_grating_cos

    return spatial_gabor_sin, spatial_gabor_cos, temporal_gabor_sin, temporal_gabor_cos


def generate_3dgabor_array(vhsize=(576,1024),
                           stimulus_fps=24,
                           aspect_ratio='auto',
                           filter_temporal_width='auto',
                           centerh=0.5,
                           centerv=0.5,
                           direction=45.0,
                           spatial_freq=16.0,
                           spatial_env=0.3,
                           temporal_freq=2.0,
                           temporal_env=0.3,
                           phase_offset=0.0):
    '''
    '''
    vdim, hdim = vhsize
    if aspect_ratio == 'auto':
        aspect_ratio = hdim/float(vdim)

    if filter_temporal_width == 'auto':
        filter_temporal_width = int(stimulus_fps*(2/3.))

    gabor_components = mk_3d_gabor(vhsize,
                                   stimulus_fps=stimulus_fps,
                                   aspect_ratio=aspect_ratio,
                                   filter_temporal_width=filter_temporal_width,
                                   centerh=centerh,
                                   centerv=centerv,
                                   direction=direction,
                                   spatial_freq=spatial_freq,
                                   spatial_env=spatial_env,
                                   temporal_freq=temporal_freq,
                                   temporal_env=temporal_env,
                                   phase_offset=phase_offset,
                                   )

    gabor_video = mk_spatiotemporal_gabor(*gabor_components)
    return gabor_video


def dotspatial_frames(spatial_gabor_sin, spatial_gabor_cos,
                      stimulus,
                      masklimit=0.001):
    '''Dot the spatial gabor filters filter with the stimulus

    Parameters
    ----------
    spatial_gabor_sin : np.array, (vdim,hdim)
    spatial_gabor_cos : np.array, (vdim,hdim)
        Spatial gabor quadrature pair
    stimulus : 2D np.array (nimages, vdim*hdim)
        The movie frames with the spatial dimension collapsed.
    masklimit : float-like
        Threshold to find the non-zero filter region

    Returns
    -------
    channel_sin : np.ndarray, (nimages, )
    channel_cos : np.ndarray, (nimages, )
        The filter response to each stimulus
        The quadrature pair can be combined: (x^2 + y^2)^0.5
    '''
    gabors = np.asarray([spatial_gabor_sin.ravel(),
                         spatial_gabor_cos.ravel()])
    # dot the gabors with the stimulus
    mask = np.abs(gabors).sum(0) > masklimit
    gabor_prod = (gabors[:,mask].squeeze() @ stimulus.T[mask].squeeze()).T
    gabor_sin, gabor_cos = gabor_prod[:,0], gabor_prod[:,1]
    return gabor_sin, gabor_cos


def dotdelay_frames(spatial_gabor_sin, spatial_gabor_cos,
                    temporal_gabor_sin, temporal_gabor_cos,
                    stimulus,
                    masklimit=0.001):
    '''Convolve the motion energy filter with a stimulus

    Parameters
    ----------
    spatial_gabor_sin : np.array, (vdim,hdim)
    spatial_gabor_cos : np.array, (vdim,hdim)
        Spatial gabor quadrature pair

    temporal_gabor_sin : np.array, (temporal_filter_width,)
    temporal_gabor_cos : np.array, (temporal_filter_width,)
        Temporal gabor quadrature pair

    stimulus : 2D np.array (nimages, vdim*hdim)
        The movie frames with the spatial dimension collapsed.

    Returns
    -------
    channel_sin : np.ndarray, (nimages, )
    channel_cos : np.ndarray, (nimages, )
        The filter response to the stimulus at each time point
        The quadrature pair can be combined: (x^2 + y^2)^0.5
    '''

    gabor_sin, gabor_cos = dotspatial_frames(spatial_gabor_sin, spatial_gabor_cos,
                                             stimulus, masklimit=masklimit)
    gabor_prod = np.c_[gabor_sin, gabor_cos]


    temporal_gabors = np.asarray([temporal_gabor_sin,
                                  temporal_gabor_cos])

    # dot the product with the temporal gabors
    outs =  gabor_prod[:, [0]] @ temporal_gabors[[1]] + gabor_prod[:, [1]] @ temporal_gabors[[0]]
    outc = -gabor_prod[:, [0]] @ temporal_gabors[[0]] + gabor_prod[:, [1]] @ temporal_gabors[[1]]

    # sum across delays
    nouts = np.zeros_like(outs)
    noutc = np.zeros_like(outc)
    tdxc = int(np.ceil(outs.shape[1]/2.0))
    delays = np.arange(outs.shape[1])-tdxc +1
    for ddx, num in enumerate(delays):
        if num == 0:
            nouts[:, ddx] = outs[:,ddx]
            noutc[:, ddx] = outc[:,ddx]
        elif num > 0:
            nouts[num:, ddx] = outs[:-num,ddx]
            noutc[num:, ddx] = outc[:-num,ddx]
        elif num < 0:
            nouts[:num, ddx] = outs[abs(num):,ddx]
            noutc[:num, ddx] = outc[abs(num):,ddx]

    channel_sin = nouts.sum(-1)
    channel_cos = noutc.sum(-1)
    return channel_sin, channel_cos


def mk_spatiotemporal_gabor(spatial_gabor_sin, spatial_gabor_cos,
                            temporal_gabor_sin, temporal_gabor_cos):
    '''Make 3D motion energy filter defined by the spatial and temporal gabors.

    Takes the output of :func:`mk_3d_gabor` and constructs the 3D filter.
    This is useful for visualization.

    Parameters
    ----------
    spatial_gabor_sin : np.array, (vdim,hdim)
    spatial_gabor_cos : np.array, (vdim,hdim)
        Spatial gabor quadrature pair
    temporal_gabor_sin : np.array, (filter_temporal_width,)
    temporal_gabor_cos : np.array, (filter_temporal_width,)
        Temporal gabor quadrature pair

    Returns
    -------
    motion_energy_filter : np.array, (vdim, hdim, filter_temporal_width)
        The motion energy filter
    '''
    a = -spatial_gabor_sin.ravel()[...,None] @ temporal_gabor_sin[...,None].T
    b =  spatial_gabor_cos.ravel()[...,None] @ temporal_gabor_cos[...,None].T
    x,y = spatial_gabor_sin.shape
    t = temporal_gabor_sin.shape[0]
    return (a+b).reshape(x,y,t)



def compute_spatial_gabor_responses(stimulus,
                                    aspect_ratio='auto',
                                    spatial_frequencies=[0,2,4,8,16,32],
                                    quadrature_combination=sqrt_sum_squares,
                                    output_nonlinearity=log_compress,
                                    dtype=np.float64,
                                    dozscore=True):
    """Compute the spatial gabor filters' response to each stimulus.

    Parameters
    ----------
    stimulus : 3D np.array (n, vdim, hdim)
        The stimulus frames.
    spatial_frequencies : array-like
        The spatial frequencies to compute. The spatial envelope is determined by this.
    quadrature_combination : function, optional
        Specifies how to combine the channel reponses quadratures.
        The function must take the sin and cos as arguments in order.
        Defaults to: (sin^2 + cos^2)^1/2
    output_nonlinearity : function, optional
        Passes the channels (after `quadrature_combination`) through a
        non-linearity. The function input is the (`n`,`nfilters`) array.
        Defaults to: ln(x + 1e-05)
    dozscore : bool, optional
        Whether to z-score the channel responses in time

    dtype : np.dtype
        Defaults to np.float64

    Returns
    -------
    filter_responses : np.array, (n, nfilters)
    """
    nimages, vdim, hdim = stimulus.shape
    vhsize = (vdim, hdim)

    if aspect_ratio == 'auto':
        aspect_ratio = hdim/float(vdim)

    stimulus = stimulus.reshape(stimulus.shape[0], -1)
    parameter_names, gabor_parameters = mk_moten_pyramid_params(
        1.,                     # fps
        filter_temporal_width=1.,
        aspect_ratio=aspect_ratio,
        temporal_frequencies=[0.],
        spatial_directions=[0.],
        spatial_frequencies=spatial_frequencies,
        )


    ngabors = gabor_parameters.shape[0]
    filters = [{name : gabor_parameters[idx, pdx] for pdx, name \
                in enumerate(parameter_names)} \
               for idx in range(ngabors)]

    info = 'Computing responses for #%i filters across #%i images (aspect_ratio=%0.03f)'
    print(info%(len(gabor_parameters), nimages, aspect_ratio))

    channels = np.zeros((nimages, len(gabor_parameters)), dtype=dtype)
    for idx, gabor_param_dict in iterator_func(enumerate(filters),
                                          '%s.compute_spatial_gabor_responses'%__name__,
                                          total=len(gabor_parameters)):
        sgabor_sin, sgabor_cos, _, _ = mk_3d_gabor(vhsize,
                                                   **gabor_param_dict)

        channel_sin, channel_cos = dotspatial_frames(sgabor_sin, sgabor_cos, stimulus)
        channel = quadrature_combination(channel_sin, channel_cos)
        channels[:, idx] = channel
    channels = output_nonlinearity(channels)
    if dozscore:
        from scipy.stats import zscore
        channels = zscore(channels)
    return channels


def compute_filter_responses(stimulus,
                             stimulus_fps,
                             aspect_ratio='auto',
                             filter_temporal_width='auto',
                             quadrature_combination=sqrt_sum_squares,
                             output_nonlinearity=log_compress,
                             dozscore=True,
                             dtype=np.float64,
                             pyramid_parameters={}):
    """Compute the motion energy filters' response to the stimuli.

    Parameters
    ----------
    stimulus : 3D np.array (n, vdim, hdim)
        The movie frames.
    stimulus_fps : scalar
        The temporal frequency of the stimulus
    aspect_ratio : bool, or scalar
        Defaults to hdim/vdim. Otherwise, pass as scalar
    filter_temporal_width : int, None
        The number of frames in one filter.
        Defaults to approximately 0.666[secs] (floor(stimulus_fps*(2/3))).
    quadrature_combination : function, optional
        Specifies how to combine the channel reponses quadratures.
        The function must take the sin and cos as arguments in order.
        Defaults to: (sin^2 + cos^2)^1/2
    output_nonlinearity : function, optional
        Passes the channels (after `quadrature_combination`) through a
        non-linearity. The function input is the (`n`,`nfilters`) array.
        Defaults to: ln(x + 1e-05)
    dozscore : bool, optional
        Whether to z-score the channel responses in time

    dtype : np.dtype
        Defaults to np.float64

    pyramid_parameters: dict
        See :func:`mk_moten_pyramid_params` for details on parameters
        specifiying a motion energy pyramid.

    Returns
    -------
    filter_responses : np.array, (n, nfilters)
    """
    nimages, vdim, hdim = stimulus.shape
    stimulus = stimulus.reshape(stimulus.shape[0], -1)
    vhsize = (vdim, hdim)

    if aspect_ratio == 'auto':
        aspect_ratio = hdim/float(vdim)

    if filter_temporal_width == 'auto':
        filter_temporal_width = int(stimulus_fps*(2./3.))

    # pass parameters
    pkwargs = dict(aspect_ratio=aspect_ratio,
                   filter_temporal_width=filter_temporal_width)
    pkwargs.update(**pyramid_parameters)
    parameter_names, gabor_parameters = mk_moten_pyramid_params(stimulus_fps,
                                                                **pkwargs)

    ngabors = gabor_parameters.shape[0]
    filters = [{name : gabor_parameters[idx, pdx] for pdx, name \
                in enumerate(parameter_names)} \
               for idx in range(ngabors)]

    info = 'Computing responses for #%i filters across #%i images (aspect_ratio=%0.03f)'
    print(info%(len(gabor_parameters), nimages, aspect_ratio))

    channels = np.zeros((nimages, len(gabor_parameters)), dtype=dtype)
    for idx, gabor_param_dict in iterator_func(enumerate(filters),
                                          '%s.compute_filter_responses'%__name__,
                                          total=len(filters)):

        gabor = mk_3d_gabor(vhsize,
                            **gabor_param_dict)

        gabor0, gabor90, tgabor0, tgabor90 = gabor

        channel_sin, channel_cos = dotdelay_frames(gabor0, gabor90,
                                                   tgabor0, tgabor90,
                                                   stimulus,
                                                   )
        channel = quadrature_combination(channel_sin, channel_cos)
        channels[:,idx] = channel
    channels = output_nonlinearity(channels)
    if dozscore:
        from scipy.stats import zscore
        channels = zscore(channels)
    return channels


def mk_moten_pyramid_params(stimulus_fps,
                            filter_temporal_width='auto',
                            aspect_ratio='auto',
                            temporal_frequencies=[0,2,4],
                            spatial_frequencies=[0,2,4,8,16,32],
                            spatial_directions=[0,45,90,135,180,225,270,315],
                            sf_gauss_ratio=0.6,
                            max_spatial_env=0.3,
                            gabor_spacing=3.5,
                            tf_gauss_ratio=10.,
                            max_temp_env=0.3,
                            spatial_phase_offset=0.0,
                            include_edges=False,
                            ):
    """Parametrize a motion energy pyramid that tiles the stimulus.

    Parameters
    ----------
    stimulus_fps : scalar, [Hz]
        Stimulus playback speed in frames per second.
    spatial_frequencies : array-like, [cycles-per-image]
        Spatial frequencies for the filters
    spatial_directions : array-like, [degrees]
        Direction of filter motion. Degree position corresponds
        to standard unit-circle coordinates (i.e. 0=right, 180=left).
    temporal_frequencies : array-like, [Hz]
        Temporal frequencies of the filters
    filter_temporal_width : int
        Temporal window of the motion energy filter (e.g. 10).
        Defaults to approximately 0.666[secs] (`floor(stimulus_fps*(2/3))`).
    aspect_ratio : optional, 'auto' or float-like,
        Defaults to stimulus aspect ratio: hdim/vdim
        Useful for preserving the spatial gabors circular even
        when images have non-square aspect ratios. For example,
        a 16:9 image would have `aspect_ratio`=16/9.

    sf_gauss_ratio : scalar
        The ratio of spatial frequency to gaussian s.d.
        This controls the number of cycles in a filter
    max_spatial_env : scalar
        Defines the maximum s.d. of the gaussian
    gabor_spacing : scalar
        Defines the spacing between spatial gabors
        (in s.d. units)
    tf_gauss_ratio : scalar
        The ratio of temporal frequency to gaussian s.d.
        This controls the number of temporal cycles
    max_temp_env : scalar
        Defines the maximum s.d. of the temporal gaussian
    include_edges : bool
        Determines whether to include filters at the edge
        of the image which might be partially outside the
        stimulus field-of-view

    Returns
    -------
    parameter_names : list of strings
        The name of the parameters
    gabor_parameters : 2D np.ndarray, (nfilters, 11)
        Parameters that define the motion energy filter
        Each of the `nfilters` has the following parameters:
            * centerv,centerh : y:vertical and x:horizontal position ('0,0' is top left)
            * direction       : direction of motion [degrees]
            * spatial_freq    : spatial frequency [cpi]
            * spatial_env     : spatial envelope (gaussian s.d.)
            * temporal_freq   : temporal frequency [Hz]
            * temporal_env    : temporal envelope (gaussian s.d.)
            * filter_temporal_width : temporal window of filter [frames]
            * aspect_ratio    : width/height
            * stimulus_fps    : stimulus playback speed in frames per second
            * spatial_phase_offset : filter phase offset in [degrees]

    Notes
    -----
    Same method as Nishimoto, et al., 2011.
    """
    assert isinstance(aspect_ratio, (int, float, np.ndarray))

    def compute_envelope(freq, ratio):
        return np.inf if freq == 0 else (1.0/freq)*ratio

    spatial_frequencies = np.asarray(spatial_frequencies)
    spatial_directions = np.asarray(spatial_directions)
    temporal_frequencies = np.asarray(temporal_frequencies)
    include_edges = int(include_edges)

    # We have to deal with zero frequency spatial filters differently
    include_local_dc = True if 0 in spatial_frequencies else False
    spatial_frequencies = np.asarray([t for t in spatial_frequencies if t != 0])

    # add temporal envelope max
    params = list(itertools.product(spatial_frequencies, spatial_directions))

    gabor_parameters = []

    for spatial_freq, spatial_direction in params:
        spatial_env = min(compute_envelope(spatial_freq, sf_gauss_ratio), max_spatial_env)

        # compute the number of gaussians that will fit in the FOV
        vertical_space = np.floor(((1.0 - spatial_env*gabor_spacing)/(gabor_spacing*spatial_env))/2.0)
        horizontal_space = np.floor(((aspect_ratio - spatial_env*gabor_spacing)/(gabor_spacing*spatial_env))/2.0)

        # include the edges of screen?
        vertical_space = max(vertical_space, 0) + include_edges
        horizontal_space = max(horizontal_space, 0) + include_edges

        # get the spatial gabor locations
        ycenters = spatial_env*gabor_spacing*np.arange(-vertical_space, vertical_space+1) + 0.5
        xcenters = spatial_env*gabor_spacing*np.arange(-horizontal_space, horizontal_space+1) + aspect_ratio/2.

        for ii, (cx, cy) in enumerate(itertools.product(xcenters,ycenters)):
            for temp_freq in temporal_frequencies:
                temp_env = min(compute_envelope(temp_freq, tf_gauss_ratio), max_temp_env)

                if temp_freq == 0 and spatial_direction >= 180:
                    # 0Hz temporal filter doesn't have motion, so
                    # 0 and 180 degrees orientations are the same filters
                    continue

                gabor_parameters.append([cx,
                                         cy,
                                         spatial_direction,
                                         spatial_freq,
                                         spatial_env,
                                         temp_freq,
                                         temp_env,
                                         filter_temporal_width,
                                         aspect_ratio,
                                         stimulus_fps,
                                         spatial_phase_offset,
                                         ])

                if spatial_direction == 0 and include_local_dc:
                    # add local 0 spatial frequency non-directional temporal filter
                    gabor_parameters.append([cx,
                                             cy,
                                             spatial_direction,
                                             0., # zero spatial freq
                                             spatial_env,
                                             temp_freq,
                                             temp_env,
                                             filter_temporal_width,
                                             aspect_ratio,
                                             stimulus_fps,
                                             spatial_phase_offset,
                                             ])

    parameter_names = ('centerh',
                       'centerv',
                       'direction',
                       'spatial_freq',
                       'spatial_env',
                       'temporal_freq',
                       'temporal_env',
                       'filter_temporal_width',
                       'aspect_ratio',
                       'stimulus_fps',
                       'spatial_phase_offset',
                       )

    gabor_parameters = np.asarray(gabor_parameters)
    return parameter_names, gabor_parameters



#!/usr/bin/env python3
import sys

try:
    import configparser
except ImportError:
    import ConfigParser as configparser


if len(set(('develop', 'bdist_wheel', 'bdist_egg',
            'bdist_rpm', 'bdist',
            'sdist', 'bdist_wheel', 'bdist_dumb',
            'bdist_wininst', 'install_egg_info',
            'egg_info', 'easy_install')).intersection(sys.argv)) > 0:

    from setuptools import setup
    from setuptools.command.install import install
else:
    # Use standard
    from distutils.core import setup
    from distutils.command.install import install

def set_default_options(optfile):
    config = configparser.ConfigParser()
    config.read(optfile)
    with open(optfile, 'w') as fp:
        config.write(fp)

class my_install(install):
    def run(self):
        install.run(self)
        optfile = [f for f in self.get_outputs() if 'defaults.cfg' in f]
        # set_default_options(optfile[0])


if not 'extra_setuptools_args' in globals():
    extra_setuptools_args = dict()


with open('README.rst', 'r') as fid:
    long_description = fid.read()


def main(**kwargs):
    setup(name='pymoten',
          version='0.0.3',
          description="""Extract motion energy features from video using spatio-temporal Gabors""",
          author='Anwar O. Nunez-Elizalde',
          author_email='anwarnunez@gmail.com',
          url='https://gallantlab.github.io/pymoten/',
          packages=['moten',
                    ],
          package_data={
              'moten':[
                'defaults.cfg',
                  ],
              },
          cmdclass=dict(install=my_install),
          include_package_data=True,
          long_description=long_description,
          long_description_content_type='text/x-rst',
          **kwargs)

if __name__ == "__main__":
    main(**extra_setuptools_args)



'''
'''
#
# Adapted from MATLAB code written by S. Nishimoto (see Nishimoto, et al., 2011).
# Anwar O. Nunez-Elizalde (Jan, 2016)
#
# Updates:
#  Anwar O. Nunez-Elizalde (Apr, 2020)
#
#
# Implementation notes:
#
# The terminology here is "pyramid" and "filters".
# In moten.core, the termilogy is inconsistent.
#
# TODO: Fix terminology in moten.core to match API.
#
import numpy as np


from moten import (utils,
                   core,
                   viz,
                   )

__all__ = ['MotionEnergyPyramid',
           'StimulusMotionEnergy',
           'StimulusStaticGaborPyramid',
           'DefaultPyramids',
           ]

##############################
#
##############################

class MotionEnergyPyramid(object):
    '''Construct a motion energy pyramid that tiles the stimulus.

    Generates motion energy filters at the desired
    spatio-temporal frequencies and directions of motion.
    Multiple motion energy filters per spatio-temporal frequency
    are constructed and each is centered at different locations
    in the image in order to tile the stimulus.

    Parameters
    ----------
    stimulus_vhsize : tuple of ints
        Size of the stimulus in pixels (vdim, hdim)
    stimulus_fps : scalar, [Hz]
        Stimulus playback speed in frames per second.
    spatial_frequencies : array-like, [cycles-per-image]
        Spatial frequencies for the filters
    spatial_directions : array-like, [degrees]
        Direction of filter motion. Degree position corresponds
        to standard unit-circle coordinates (i.e. 0=right, 180=left).
    temporal_frequencies : array-like, [Hz]
        Temporal frequencies of the filters
    filter_temporal_width : int
        Temporal window of the motion energy filter (e.g. 10).
        Defaults to approximately 0.666[secs] (`floor(stimulus_fps*(2/3))`).

    Attributes
    ----------
    nfilters : int
    filters : list of dicts
        Each item is a dict defining a motion energy filter.
        Each of the `nfilters` has the following parameters:
            * centerv,centerh : y:vertical and x:horizontal position ('0,0' is top left)
            * direction       : direction of motion [degrees]
            * spatial_freq    : spatial frequency [cpi]
            * spatial_env     : spatial envelope (gaussian s.d.)
            * temporal_freq   : temporal frequency [Hz]
            * temporal_env    : temporal envelope (gaussian s.d.)
            * filter_temporal_width : temporal window of filter [frames]
            * aspect_ratio    : width/height
            * stimulus_fps    : stimulus playback speed in frames per second
            * spatial_phase_offset : filter phase offset in [degrees]

    parameters : dict of arrays
        The individual parameter values across all filters.
        Each item is an array of length ``nfilters``
    definition : dict
        Parameters used to define the pyramid.
    parameters_matrix : np.array, (nfilters, 11)
        Parameters that defined the motion energy filter.
    parameters_names  : tuple of strings

    Notes
    -----
    See :func:`moten.core.mk_moten_pyramid_params` for more details on
    making the pyramid.
    '''
    def __init__(self,
                 stimulus_vhsize=(576, 1024),
                 stimulus_fps=24,
                 temporal_frequencies=[0,2,4],
                 spatial_frequencies=[0,2,4,8,16],
                 spatial_directions=[0,45,90,135,180,225,270,315],
                 sf_gauss_ratio=0.6,
                 max_spatial_env=0.3,
                 filter_spacing=3.5,
                 tf_gauss_ratio=10.,
                 max_temp_env=0.3,
                 include_edges=False,
                 spatial_phase_offset=0.0,
                 filter_temporal_width='auto'):
        '''...More parameters

        Parameters
        ----------
        sf_gauss_ratio : scalar
            The ratio of spatial frequency to gaussian s.d.
            This controls the number of cycles in a filter
        max_spatial_env : scalar
            Defines the maximum s.d. of the gaussian
        gabor_spacing : scalar
            Defines the spacing between spatial gabors
            (in s.d. units)
        tf_gauss_ratio : scalar
            The ratio of temporal frequency to gaussian s.d.
            This controls the number of temporal cycles
        max_temp_env : scalar
            Defines the maximum s.d. of the temporal gaussian
        include_edges : bool
            Determines whether to include filters at the edge
            of the image which might be partially outside the
            stimulus field-of-view
        '''
        vdim, hdim = stimulus_vhsize
        aspect_ratio = hdim/float(vdim) # same as stimulus

        if filter_temporal_width == 'auto':
            # default to 2/3 stimulus frame rate
            filter_temporal_width = int(stimulus_fps*(2/3.))

        stimulus_vht_fov = (vdim, hdim, stimulus_fps)
        filter_vht_fov = (vdim, hdim, filter_temporal_width)

        # store pyramid parameters
        definition = utils.DotDict()
        for local_name, local_values in sorted(locals().items()):
            if local_name in ('self', 'definition') :
                continue
            definition[local_name] = local_values

        # sanity checks
        assert isinstance(stimulus_fps, int)
        assert isinstance(filter_temporal_width, int)

        # construct the gabor pyramid
        parameter_names, filter_params_array = core.mk_moten_pyramid_params(
            stimulus_fps,
            filter_temporal_width=filter_temporal_width,
            aspect_ratio=aspect_ratio,
            temporal_frequencies=temporal_frequencies,
            spatial_frequencies=spatial_frequencies,
            spatial_directions=spatial_directions,
            sf_gauss_ratio=sf_gauss_ratio,
            max_spatial_env=max_spatial_env,
            gabor_spacing=filter_spacing,
            tf_gauss_ratio=tf_gauss_ratio,
            max_temp_env=max_temp_env,
            include_edges=include_edges,
            spatial_phase_offset=spatial_phase_offset,
            )

        nfilters = filter_params_array.shape[0]

        # make individual filters readable
        filters = [utils.DotDict({name : filter_params_array[idx, pdx] for pdx, name \
                                  in enumerate(parameter_names)}) \
                   for idx in range(nfilters)]

        # parameters across filters
        parameters = utils.DotDict({name : filter_params_array[:,idx] for idx, name \
                                    in enumerate(parameter_names)})

        # store pyramid information as attributes
        self.filters = filters
        self.nfilters = nfilters
        self.parameters = parameters
        self.definition = definition
        self.parameters_matrix = filter_params_array
        self.parameters_names = parameter_names
        self.mask_threshold = 0.001 # HARDCODED. TODO: Make parameter.

    def __repr__(self):
        info = '<%s.%s [#%i filters (ntfq=%i, nsfq=%i, ndirs=%i) aspect=%0.03f]>'
        details = (__name__, type(self).__name__, self.nfilters,
                   len(self.definition.temporal_frequencies),
                   len(self.definition.spatial_frequencies),
                   len(self.definition.spatial_directions),
                   self.definition.aspect_ratio)
        return info%details

    def filters_at_vhposition(self, centerv, centerh):
        '''Center spatio-temporal filters to requested vh-position.

        Parameters
        ----------
        centerv : scalar
            Vertical filter position from top of frame (min=0, max=1.0).
        centerh : scalar
            Horizontal filter position from left of frame (min=0, max=aspect_ratio).

        Returns
        -------
        centered_filters : list of dicts
            Spatio-temporal filter parameters at vh-position.
        '''
        unique_parameters = np.unique(self.parameters_matrix[:, 2:], axis=0)
        nunique = unique_parameters.shape[0]

        new_filters = [utils.DotDict({name : unique_parameters[idx, pdx] for pdx, name \
                                      in enumerate(self.parameters_names[2:])}) \
                       for idx in range(nunique)]
        for filt in new_filters:
            filt['centerh'] = centerh
            filt['centerv'] = centerv
        return new_filters

    def get_filter_spatiotemporal_quadratures(self, filterid=0):
        '''Generate the spatial and temporal arrays that define the motion energy filter.

        Parameters
        ----------
        filterid : int, or dict
            If int, the filter index.
            If dict, a filter dictionary definition

        Returns
        -------
        spatial_gabor_sin : 2D np.ndarray, (vdim, hdim)
        spatial_gabor_cos : 2D np.ndarray, (vdim, hdim)
            Spatial gabor quadrature pair. ``spatial_gabor_cos`` has
            a 90 degree phase offset relative to ``spatial_gabor_sin``

        temporal_gabor_sin : 1D np.ndarray, (`filter_temporal_width`,)
        temporal_gabor_cos : 1D np.ndarray, (`filter_temporal_width`,)
            Temporal gabor quadrature pair. ``temporal_gabor_cos`` has
            a 90 degree phase offset relative to ``temporal_gabor_sin``
        '''
        vhsize = (self.definition.vdim, self.definition.hdim)

        # extract parameters for this filterid
        if isinstance(filterid, dict):
            gabor_parameters_dict = filterid.copy()
        else:
            gabor_parameters_dict = self.filters[filterid]

        sgabor0, sgabor90, tgabor0, tgabor90 = core.mk_3d_gabor(vhsize,
                                                                **gabor_parameters_dict)
        return sgabor0, sgabor90, tgabor0, tgabor90

    def get_filter_temporal_quadrature(self, filterid=0):
        '''Generate the temporal arrays that define the motion energy filter.

        Parameters
        ----------
        filterid : int, or dict
            If int, the filter index.
            If dict, a filter dictionary definition

        Returns
        -------
        temporal_gabor_sin : 1D np.ndarray, (`filter_temporal_width`,)
        temporal_gabor_cos : 1D np.ndarray, (`filter_temporal_width`,)
            Temporal gabor quadrature pair. ``temporal_gabor_cos`` has
            a 90 degree phase offset relative to ``temporal_gabor_sin``
        '''
        _,_, tgabor0, tgabor90 = self.get_filter_spatiotemporal_quadratures(filterid=filterid)
        return tgabor0, tgabor90

    def get_filter_spatial_quadrature(self, filterid=0):
        '''Generate the spatial arrays that define the motion energy filter.

        Parameters
        ----------
        filterid : int, or dict
            If int, the filter index.
            If dict, a filter dictionary definition

        Returns
        -------
        spatial_gabor_sin : 2D np.ndarray, (vdim, hdim)
        spatial_gabor_cos : 2D np.ndarray, (vdim, hdim)
            Spatial gabor quadrature pair. ``spatial_gabor_cos`` has
            a 90 degree phase offset relative to ``spatial_gabor_sin``
        '''
        sgabor0, sgabor90, _, _ = self.get_filter_spatiotemporal_quadratures(filterid=filterid)
        return sgabor0, sgabor90

    def get_filter_mask(self, filterid=0):
        '''Generate a mask of the filter in the image

        Parameters
        ----------
        filterid : int, or dict
            If int, the filter index.
            If dict, a filter dictionary definition

        Returns
        -------
        mask : 2D np.ndarray, (vdim, hdim)
            Filter spatial mask
        '''
        abs = np.abs
        threshold = self.mask_threshold

        spsin, spcos = self.get_filter_spatial_quadrature(filterid)
        mask = (abs(spsin) + abs(spcos)) > threshold
        return mask

    def get_filter_pixel_sizes(self):
        '''Measure the size of the each filter in the pyramid in pixels.

        Returns
        -------
        filter_npixels : 2D np.ndarray, (nfilters,)
            Array containing the size of each filter in pixels.
        '''
        npixels = []
        for filter_idx in range(self.nfilters):
            mask = self.get_filter_mask(filter_idx)
            npixels.append(mask.sum())
        return npixels


    def show_filter(self, filterid=0, speed=1.0, background=None):
        '''Display the motion energy filter as an animation.

        Parameters
        ----------
        filterid : int, or dict
            If int, it's the index of the filter to display.
            If dict, it's a motion energy filter definition.

        Returns
        -------
        animation : matplotlib.animation

        Examples
        --------
        >>> import moten
        >>> pyramid = moten.get_default_pyramid()
        >>> _ = pyramid.show_filter(12)
        >>> custom_filter = {'centerh': 0.8888888888888888,
                             'centerv': 0.5,
                             'direction': 45.0,
                             'spatial_freq': 2.0,
                             'spatial_env': 0.3,
                             'temporal_freq': 2.0,
                             'temporal_env': 0.3,
                             'filter_temporal_width': 16.0,
                             'aspect_ratio': 1.7777777777777777,
                             'stimulus_fps': 24.0,
                             'spatial_phase_offset': 0.0}
        >>> _ = pyramid.show_filter(custom_filter)
        '''
        # Get dimensions of movie
        vdim, hdim, stimulus_fps = self.definition.stimulus_vht_fov
        aspect_ratio = self.definition.aspect_ratio
        filter_width = self.definition.filter_temporal_width
        vhsize = (vdim, hdim)

        # extract parameters for this filterid
        if isinstance(filterid, dict):
            gabor_params_dict = filterid.copy()
            filterid = 0
        else:
            gabor_params_dict = self.filters[filterid]

        # construct title from parameters
        title = ''
        for pdx, (pname, pval) in enumerate(sorted(gabor_params_dict.items())):
            title += '%s=%0.02f, '%(pname, pval)
            if np.mod(pdx, 3) == 0 and pdx > 0:
                title += '\n'

        return viz.plot_3dgabor(vhsize,
                                gabor_params_dict,
                                background=background,
                                title=title, speed=speed)

    def project_stimulus(self,
                         stimulus,
                         filters='all',
                         quadrature_combination=utils.sqrt_sum_squares,
                         output_nonlinearity=utils.log_compress,
                         dtype='float32',
                         use_cuda=False):
        '''Compute the motion energy filter responses to the stimulus.

        Parameters
        ----------
        stimulus : np.ndarray, (nimages, vdim, hdim) or (nimages, npixels)
            The movie frames.
            If ``stimulus`` is two-dimensional with shape (nimages, npixels), then
            ``vhsize=(vdim,hdim)`` is required and `npixels == vdim*hdim`.
        filters : optional, 'all' or list of dicts
            By default compute the responses for all filters.
            Otherwise, provide a list of filter definitions to use.
        quadrature_combination : function, optional
            Specifies how to combine the channel reponse quadratures.
            The function must take the sin and cos as arguments in that order.
            Defaults to: :math:`(sin^2 + cos^2)^{1/2}`
        output_nonlinearity : function, optional
            Passes the channels (after ``quadrature_combination``) through a
            non-linearity. The function input is the (`nimages, nfilters`) array.
            Defaults to: :math:`log(x + 1e-05)`

        Returns
        -------
        filter_responses : np.ndarray, (nimages, nfilters)
        '''
        if filters == 'all':
            filters = self.filters

        vdim, hdim = self.definition.stimulus_vhsize

        output = core.project_stimulus(stimulus,
                                       filters,
                                       quadrature_combination=quadrature_combination,
                                       output_nonlinearity=output_nonlinearity,
                                       vhsize=(vdim, hdim),
                                       dtype=dtype)

        return output

    def raw_project_stimulus(self, stimulus, filters='all', dtype='float32'):
        '''Obtain responses to the stimulus from all filter quadrature-pairs.

        Parameters
        ----------
        stimulus : np.array, (nimages, vdim, hdim)
            The movie frames.
        filters : optional, 'all' or list of dicts
            By default compute the responses for all filters.
            Otherwise, provide a list of filter definitions to use.

        Returns
        -------
        output_sin : np.ndarray, (nimages, nfilters)
        output_cos : np.ndarray, (nimages, nfilters)
        '''
        if filters == 'all':
            filters = self.filters

        vdim, hdim = self.definition.stimulus_vhsize

        output_sin, output_cos = core.raw_project_stimulus(stimulus,
                                                           filters,
                                                           vhsize=(vdim, hdim),
                                                           dtype=dtype)

        return output_sin, output_cos


class StimulusMotionEnergy(object):
    '''Parametrize a motion energy pyramid that tiles the stimulus.

    Generates motion energy filters at the desired
    spatio-temporal frequencies and directions of motion.
    Multiple motion energy filters per spatio-temporal frequency
    are constructed and each is centered at different locations
    in the image in order to tile the stimulus.

    Parameters
    ----------
    stimulus : np.array, (nimages, vdim, hdim)
        The movie frames.
    stimulus_fps : scalar, [Hz]
        Stimulus playback speed in frames per second.
    spatial_frequencies : array-like, [cycles-per-image]
        Spatial frequencies for the filters
    spatial_directions : array-like, [degrees]
        Direction of filter motion. Degree position corresponds
        to standard unit-circle coordinates (i.e. 0=right, 180=left).
    temporal_frequencies : array-like, [Hz]
        Temporal frequencies of the filters
    filter_temporal_width : int
        Temporal window of the motion energy filter (e.g. 10).
        Defaults to approximately 0.666[secs] (`floor(stimulus_fps*(2/3))`).

    Attributes
    ----------
    stimulus : 2D np.ndarray, (nimages, vdim*hdim)
        Time-space representation of stimulus
    view : :class:`MotionEnergyPyramid`
        Full description of the motion energy pyramid.
    nimages : int
        Number of video frames
    nfilters : int
    aspect_ratio : scalar, (hdim/vdim)
    stimulus_fps : int (fps)
    stimulus_vhsize : tuple of ints, (vdim, hdim)
    original_stimulus : 3D np.ndarray, (nimages, vdim, hdim)
    '''
    def __init__(self,
                 stimulus,
                 stimulus_fps,
                 temporal_frequencies=[0,2,4],
                 spatial_frequencies=[0,2,4,8,16],
                 spatial_directions=[0,45,90,135,180,225,270,315],
                 sf_gauss_ratio=0.6,
                 max_spatial_env=0.3,
                 filter_spacing=3.5,
                 tf_gauss_ratio=10.,
                 max_temp_env=0.3,
                 include_edges=False,
                 spatial_phase_offset=0.0,
                 filter_temporal_width='auto'):
        '''
        Notes
        -----
        See :class:`MotionEnergyPyramid` for more detail on
        keyword arguments used to construct the motion energy pyramid.
        '''
        nimages, vdim, hdim = stimulus.shape
        stimulus_vhsize = (vdim, hdim)
        original_stimulus = stimulus
        stimulus = stimulus.reshape(stimulus.shape[0], -1)
        aspect_ratio = hdim/vdim

        pyramid = MotionEnergyPyramid(stimulus_vhsize=(vdim, hdim),
                                      stimulus_fps=stimulus_fps,
                                      filter_temporal_width=filter_temporal_width,
                                      temporal_frequencies=temporal_frequencies,
                                      spatial_frequencies=spatial_frequencies,
                                      spatial_directions=spatial_directions,
                                      sf_gauss_ratio=sf_gauss_ratio,
                                      max_spatial_env=max_spatial_env,
                                      filter_spacing=filter_spacing,
                                      tf_gauss_ratio=tf_gauss_ratio,
                                      max_temp_env=max_temp_env,
                                      spatial_phase_offset=spatial_phase_offset,
                                      include_edges=include_edges)

        self.nimages = nimages
        self.stimulus = stimulus
        self.stimulus_fps = stimulus_fps
        self.stimulus_vhsize = stimulus_vhsize
        self.original_stimulus = original_stimulus
        self.aspect_ratio = aspect_ratio

        self.view = pyramid
        self.nfilters = pyramid.nfilters

    def __repr__(self):
        info = '<%s.%s [%i nframes @ %i fps. #%i filters. aspect=%0.03f]>'
        details = (__name__, type(self).__name__,
                   self.nimages,
                   self.stimulus_fps,
                   self.nfilters,
                   self.aspect_ratio)
        return info%details

    def project_stimulus(self, *args, **kwargs):
        '''See :meth:`MotionEnergyPyramid.project_stimulus`
        '''
        filter_responses  = self.view.project_stimulus(*args, **kwargs)
        return filter_responses

    def project(self, filters='all',
                quadrature_combination=utils.sqrt_sum_squares,
                output_nonlinearity=utils.log_compress,
                dtype='float32'):
        '''Compute the motion energy filter responses to the stimulus.

        Parameters
        ----------
        filters : optional, 'all' or list of dicts
            By default compute the responses for all filters.
            Otherwise, provide a list of filter definitions to use.
        quadrature_combination : function, optional
            Specifies how to combine the channel reponse quadratures.
            The function must take the sin and cos as arguments in that order.
            Defaults to: :math:`(sin^2 + cos^2)^{1/2}`
        output_nonlinearity : function, optional
            Passes the channels (after ``quadrature_combination``) through a
            non-linearity. The function input is the (`nimages, nfilters`) array.
            Defaults to: :math:`log(x + 1e-05)`

        Returns
        -------
        filter_responses : np.ndarray, (nimages, nfilters)
        '''
        if filters == 'all':
            filters = self.view.filters
        stimulus = self.original_stimulus

        filter_responses = self.view.project_stimulus(
            stimulus,
            filters=filters,
            quadrature_combination=quadrature_combination,
            output_nonlinearity=output_nonlinearity,
            dtype=dtype)

        return filter_responses

    def project_at_vhposition(self, centerv, centerh,
                              quadrature_combination=utils.sqrt_sum_squares,
                              output_nonlinearity=utils.log_compress,
                              dtype='float32'):
        '''Center filters at vh-position and compute their response to the stimulus.

        Parameters
        ----------
        centerv : scalar
            Vertical filter from top of frame (min=0, max=1.0).
        centerh : scalar
            Horizontal filter position from left of frame (min=0, max=aspect_ratio).

        Returns
        -------
        centered_filters : list of dicts
            Spatio-temporal filter parameters at vh-position.
        filter_responses : np.ndarray, (nimages, len(vh_centered_filters))
        '''
        # center all spatial and temporal filters at desired vh-position
        filters = self.view.filters_at_vhposition(centerv, centerh)
        stimulus = self.original_stimulus

        filter_responses = self.view.project_stimulus(
            stimulus,
            filters=filters,
            quadrature_combination=quadrature_combination,
            output_nonlinearity=output_nonlinearity,
            dtype=dtype)

        return filters, filter_responses

    def raw_projection(self, filters='all', dtype='float32'):
        '''Obtain responses to the stimulus from all filter quadrature-pairs.

        Parameters
        ----------
        filters : optional, 'all' or list of dicts
            By default compute the responses for all filters.
            Otherwise, provide a list of filter definitions to use.

        Returns
        -------
        output_sin : np.ndarray, (nimages, nfilters)
        output_cos : np.ndarray, (nimages, nfilters)
        '''
        if filters == 'all':
            filters = self.view.filters
        stimulus = self.original_stimulus

        output_sin, output_cos = self.view.raw_project_stimulus(
            stimulus, filters=filters, dtype=dtype)

        return output_sin, output_cos


class StimulusStaticGaborPyramid(StimulusMotionEnergy):
    '''
    '''
    def __init__(self,
                 stimulus,
                 spatial_frequencies=[0,2,4,8,16],
                 spatial_orientations=(0,45,90,135),
                 sf_gauss_ratio=0.6,
                 max_spatial_env=0.3,
                 filter_spacing=3.5,
                 include_edges=False,
                 spatial_phase_offset=0.0,
                 ):
        super(type(self), self).__init__(stimulus,
                                         spatial_frequencies=spatial_frequencies,
                                         spatial_directions=spatial_orientations,
                                         spatial_phase_offset=spatial_phase_offset,
                                         include_edges=include_edges,
                                         sf_gauss_ratio=sf_gauss_ratio,
                                         max_spatial_env=max_spatial_env,
                                         filter_spacing=filter_spacing,
                                         # fixed parameters for static filters:
                                         max_temp_env=np.inf,
                                         stimulus_fps=1,
                                         temporal_frequencies=[0],
                                         filter_temporal_width=1,
                                         tf_gauss_ratio=10)


class DefaultPyramids(object):
    '''
    '''
    def __init__(self):
        pass

    @property
    def pyramid15fps512x512(self):
        pyramid =  MotionEnergyPyramid(stimulus_vhsize=(512, 512),
                                       stimulus_fps=15)
        return pyramid

    @property
    def pyramid15fps96x96(self):
        pyramid =  MotionEnergyPyramid(stimulus_vhsize=(96, 96),
                                       stimulus_fps=15)
        return pyramid

    @property
    def pyramid24fps800x600(self):
        pyramid =  MotionEnergyPyramid(stimulus_vhsize=(600, 800),
                                       stimulus_fps=24)
        return pyramid

    @property
    def pyramid24fps640x480(self):
        pyramid =  MotionEnergyPyramid(stimulus_vhsize=(480, 640),
                                       stimulus_fps=24)
        return pyramid

    @property
    def pyramid24fps128x72(self):
        pyramid =  MotionEnergyPyramid(stimulus_vhsize=(72, 128),
                                       stimulus_fps=24)
        return pyramid

    @property
    def pyramid24fps256x144(self):
        pyramid =  MotionEnergyPyramid(stimulus_vhsize=(144, 256),
                                       stimulus_fps=24)
        return pyramid
